# Copyright (c) 2002 Infrae. All rights reserved.
# See also LICENSE.txt
# $Revision: 1.21 $
import Globals
from AccessControl import ClassSecurityInfo
from OFS.SimpleItem import SimpleItem
from Products.PageTemplates.PageTemplateFile import PageTemplateFile
from helpers import add_and_edit
from Persistence import Persistent

class EditorServiceError(Exception):
    pass

class EditorSession(Persistent):
    def __init__(self):
        self.document_context = {
            'depth': 0,
            'wr_name': None,
            'node_path': None,
            'widget_path': None,
            }
        self.contexts = {}
        self.cache = {}
        self.roots = []
        
    # MANIPULATORS
    def setDocumentContext(self, wr_name):
        self.document_context['wr_name'] = wr_name
        self.document_context = self.document_context
        
    def setContext(self, node, wr_name):
        self.clearNodeWidget(node)
        self.contexts[node.getNodePath('widget')] = {
            'depth': self.getContext(node)['depth'] + 1,
            'wr_name': wr_name,
            'node_path': None,
            'widget_path': None
            }
        self.contexts = self.contexts
        
    def setNodeWidget(self, node, widget_path):
        self.clearNodeWidget(node)
        context = self.getContext(node)
        context['node_path'] = node.getNodePath('widget')
        context['widget_path'] = widget_path
        self.document_context = self.document_context
        self.contexts = self.contexts
        
    def clearNodeWidget(self, node):
        context = self.getContext(node)
        context['node_path'] = None
        context['widget_path'] = None
        d = context['depth']
        for node_path, context in self.contexts.items():
            if context['depth'] > d:
                del self.contexts[node_path]
        self.contexts = self.contexts
                
    def clearContext(self, node):
        while node is not None:
            node_path = node.getNodePath('widget')
            context = self.contexts.get(node_path)
            if context is not None:
                del self.contexts[node_path]
                self.contexts = self.contexts
                return
            node = node.parentNode
    
    def pushRoot(self, node, wr_name):
        self.setContext(node, wr_name)
        self.roots.append(node.getNodePath('widget'))
        self.roots = self.roots

    def popRoot(self, node):
        self.clearContext(node)
        self.roots.pop()
        self.roots = self.roots
        
    # ACCESSORS
    def getContext(self, node):
        if not self.contexts:
            return self.document_context
        while node is not None:
            context = self.contexts.get(node.getNodePath('widget'))
            if context is not None:
                return context
            node = node.parentNode
        return self.document_context

    def getRoot(self):
        if self.roots:
            return self.roots[-1]
        else:
            return None
        
class EditorService(SimpleItem):
    """A user using an editor uses the editor session (which uses
    Zope's sessions).
    """

    security = ClassSecurityInfo()
    
    meta_type = "XMLWidgets Editor Service"
    
    manage_options = (
        {'label':'Info', 'action':'info_tab'},
        ) + SimpleItem.manage_options

    info_tab = PageTemplateFile(
        'www/editorServiceInfoTab',
        globals(),  __name__='info_tab')

    def __init__(self, id, title):
        self.id = id
        self.title = title
        
    # MANIPULATORS
    security.declareProtected('Use XMLWidgets Editor Service',
                              'setDocumentEditor')
    def setDocumentEditor(self, node, wr_name):
        self._createEditorSession(node)
        self._getEditorSession(node).setDocumentContext(wr_name)

    security.declareProtected('Use XMLWidgets Editor Service',
                              'setEditor')    
    def setEditor(self, node, wr_name):
        self._getEditorSession(node).setContext(node, wr_name)

    security.declareProtected('View',
                              'setViewer')
    def setViewer(self, wr_name):
        self.REQUEST.set('xmlwidgets_viewer', wr_name)
        
    security.declareProtected('Use XMLWidgets Editor Service',
                              'clearEditor')
    def clearEditor(self, node):
        self._getEditorSession(node).clearContext(node)
        
    security.declareProtected('Use XMLWidgets Editor Service',
                              'setNodeWidget')
    def setNodeWidget(self, node, widget_path):
        self._getEditorSession(node).setNodeWidget(node, widget_path)
        
    security.declareProtected('Use XMLWidgets Editor Service',
                              'clearNodeWidget') 
    def clearNodeWidget(self, node):
        self._getEditorSession(node).clearNodeWidget(node)

    security.declareProtected('Use XMLWidgets Editor Service',
                              'pushRoot')
    def pushRoot(self, node, wr_name):
        self._getEditorSession(node).pushRoot(node, wr_name)

    security.declareProtected('Use XMLWidgets Editor Service',
                              'popRoot')
    def popRoot(self, node):
        self._getEditorSession(node).popRoot(node)

    # ACCESSORS

    security.declareProtected('Use XMLWidgets Editor Service',
                              'getRoot')
    def getRoot(self, node):
        document = node.ownerDocument
        editor_session = self._getEditorSession(node)
        if editor_session is None:
            return document.documentElement
        root = editor_session.getRoot()
        if root is not None:
            return document.resolveNodePath(root)
        else:
            return document.documentElement

    security.declareProtected('View', 'getViewer')
    def getViewer(self):
        return getattr(self, self.REQUEST['xmlwidgets_viewer'])

    security.declareProtected('Use XMLWidgets Editor Service',
                              'getNode')
    def getNode(self, node):
        """Get the highlighted node in context.
        """
        node_path = self._getEditorSession(node).getContext(None)['node_path']
        if node_path is not None:
            to_ret = node.ownerDocument.resolveNodePath(node_path)
            return to_ret
        else:
            return None
        
    security.declareProtected('Use XMLWidgets Editor Service',
                              'render')
    def render(self, node):
        return self.getWidget(node).render()

    security.declareProtected('View', 'renderView')
    def renderView(self, node):
        # NOTE: Need to store viewer as it may change otherwise during rendering
        viewer = self.getViewer()
        return viewer.getWidget(node).render()
				
    security.declareProtected('Use XMLWidgets Editor Service',
                              'renderElements')
    def renderElements(self, node):
        return ''.join([self.render(child) for child in node.childNodes if child.nodeType == node.ELEMENT_NODE])

    security.declareProtected('View', 'renderElementsView')
    def renderElementsView(self, node):
        # NOTE: Need to store viewer as it may change otherwise during rendering
        viewer = self.getViewer()
        return ''.join([viewer.getWidget(child).render() for child in node.childNodes if child.nodeType == node.ELEMENT_NODE])

    security.declareProtected('Use XMLWidgets Editor Service',
                              'renderCache')
    def renderCache(self, node):
        editor_session = self._getEditorSession(node)
        node_path = node.getNodePath('widget')
        widget_path = self._getWidgetPath(node)
        data = editor_session.cache.get((node_path, widget_path))
        if data is None:
            widget = self._resolveWidgetPath(widget_path)
            data = widget.render()
            if getattr(widget, 'cached', lambda: 0)():
                #print "Caching:", widget_path, node_path
                editor_session.cache[(node_path, widget_path)] = data
                editor_session.cache = editor_session.cache
            #else:
            #    print "Not caching:", widget_path, node_path
        #else:
        #    print "Retrieving:", widget_path, node_path
        return data
    
    security.declareProtected('Use XMLWidgets Editor Service',
                              'renderElementsCache')
    def renderElementsCache(self, node):
        return ''.join([self.renderCache(child) for child in node.childNodes if child.nodeType == node.ELEMENT_NODE])

    security.declareProtected('Use XMLWidgets Editor Service',
                              'invalidateWidgetCache')
    def invalidateWidgetCache(self, node, widget_path):
        """Invalidates cache for 'modeless' widgets"""
        request = self.REQUEST
        r_node = request.node
        r_wr_name = request.wr_name
        node_path = node.getNodePath('widget')
        editor_session = self._getEditorSession(node)
        try:
            del editor_session.cache[(node_path, widget_path)]
        except KeyError:
            pass            
        request['node'] = r_node
        request.wr_name = r_wr_name
        editor_session.cache = editor_session.cache
        
    security.declareProtected('Use XMLWidgets Editor Service',
                              'invalidateWidgetCaches')
    def invalidateWidgetCaches(self, node, widget_path, modes):
        request = self.REQUEST
        r_node = request.node
        r_wr_name = request.wr_name
        node_path = node.getNodePath('widget')
        editor_session = self._getEditorSession(node)
        for mode in modes:
            try:
                path = widget_path + (mode,)
                del editor_session.cache[(node_path, path)]
                #print "Invalidating:", path, node_path
            except KeyError:
                pass            
        request['node'] = r_node
        request.wr_name = r_wr_name
        editor_session.cache = editor_session.cache
        
    security.declareProtected('Use XMLWidgets Editor Service',
                              'invalidateCaches')
    def invalidateCaches(self, node, modes):
        request = self.REQUEST
        r_node = request.node
        r_wr_name = request.wr_name
        widget_path = self._getWidgetPath(node)[:-1]
        node_path = node.getNodePath('widget')
        editor_session = self._getEditorSession(node)
        for mode in modes:
            try:
                path = widget_path + (mode,)
                del editor_session.cache[(node_path, path)]
            except KeyError:
                pass            
        request['node'] = r_node
        request.wr_name = r_wr_name
        editor_session.cache = editor_session.cache
        
    security.declareProtected('Use XMLWidgets Editor Service',
                              'invalidateChildCaches')
    def invalidateChildCaches(self, node, modes):
        """Invalidate cache of children (one level)
        """
        for child in node.childNodes:
            if child.nodeType != node.ELEMENT_NODE:
                continue
            self.invalidateCaches(child, modes)
    
    security.declareProtected('Use XMLWidgets Editor Service',
                              'clearCache')    
    def clearCache(self, node):
        self._getEditorSession(node).cache = {}

    security.declareProtected('Use XMLWidgets Editor Service',
                              'getWidget')
    def getWidget(self, node):
        widget_path = self._getWidgetPath(node)
        if widget_path is None:
            raise EditorServiceError, "Unknown widget for node: %s" % node.nodeName
        return self._resolveWidgetPath(widget_path)

    security.declareProtected('Use XMLWidgets Editor Service',
                              'hasNodeWidget')
    def hasNodeWidget(self, node):
        context = self._getEditorSession(node).getContext(node)
        return context['node_path'] is not None
    
    security.declareProtected('Use XMLWidgets Editor Service',
                              'getNodeWidgetPath')
    def getNodeWidgetPath(self, node):
        editor_session = self._getEditorSession(node)
        if not editor_session:
            return None
        return editor_session.getContext(node)['widget_path']
    
    # HELPERS

    def _getWidgetPath(self, node):
        context = self._getEditorSession(node).getContext(node)
        self.REQUEST['node'] = node
        self.REQUEST.wr_name = context['wr_name']
        if (context['node_path'] and
            node.getNodePath('widget') == context['node_path']):
            return context['widget_path']
        else:
            return getattr(self, context['wr_name']).findWidget(node)
        
    def _resolveWidgetPath(self, widget_path):
        obj = self.aq_inner
        for step in widget_path:
            obj = getattr(obj, step)
        return obj

    def _createEditorSession(self, node):
        session = self.REQUEST.SESSION
        session_data = session.get('xmlwidgets_service_editor', {})
        document_path =  node.ownerDocument.getPhysicalPath()
        if session_data.has_key(document_path):
            return
        editor_session = session_data[document_path] = EditorSession()
        session['xmlwidgets_service_editor'] = session_data
        
    def _getEditorSession(self, node):
        try:
            return self.REQUEST.SESSION['xmlwidgets_service_editor'][node.ownerDocument.getPhysicalPath()]
        except KeyError:
            return None

Globals.InitializeClass(EditorService)

manage_addEditorServiceForm = PageTemplateFile(
    "www/editorServiceAdd", globals(),
    __name__='manage_addEditorServiceForm')

def manage_addEditorService(self, id, title="", REQUEST=None):
    """Add Editor service to folder.
    """
    # add actual object
    id = self._setObject(id, EditorService(id, title))
    # respond to the add_and_edit button if necessary
    add_and_edit(self, id, REQUEST)
    return ''

